'''
A resolver that searches a list of include paths for YAML files matching the
given name. The include paths must be set prior to validation.
work.
'''

import functools
import pkg_resources
import os

import yaml


HANDLER_NAME = 'include'


def _handler(include_paths, uri):
    '''Handles an URI of form include://'''
    base = 'include://'
    assert(uri.startswith(base))
    uri = uri[len(base):]

    data = None
    for path in include_paths:
        uri = os.path.join(path, uri)
        try:
            f = open(uri, 'r')
        except FileNotFoundError:
            continue
        else:
            with f:
                data = yaml.load(f)
                break
    return data


def make_handler(include_paths):
    '''Returns an include handler and name (for use with make_resolver) that
    uses the give include paths.'''
    return (functools.partial(_handler, include_paths), HANDLER_NAME)


def make_default_handler():
    '''Returns a default handler, which uses the common schemas defined in
    sschema as the include path.'''
    schema_path = pkg_resources.resource_filename('sschema', 'schema')
    return make_handler([schema_path])



'''
A format checker that checks that the following for an object:
  - Each value has a "offset" and "len" integer key. "len" is also allowed to
    be a string, as long as the string is "all", indicating that this key is
    the only one.
  - Taken together, the "offset" and "len" keys carve out a contiguous set with
    no gaps, with the minimum entry being 0.
One example use of this format is in schemas describing offsets to bytes in
memory.

For example, the following is valid:
a:
  offset: 4
  len: 8
b:
  offset: 0
  len: 4
c:
  offset: 12
  len: 20

But the following is not valid:
a:
  offset: 4
  len: 8
b:
  offset: 0
  len: 5
c:
  offset: 11
  len: 20
'''


import jsonschema


FORMAT_NAME = 'contiguous-indices'


def check(d):
    '''Scans through a map of schema keys, adding up the offset and length
    properties and making sure the entire block is contiguous.'''
    if not isinstance(d, dict):
        raise jsonschema.exceptions.FormatError('not a map: %s' % d)

    # Validate offset properties for correctness.
    items = list(d.items())
    for k, v in items:
        try:
            offset = v['offset']
        except KeyError:
            raise jsonschema.exceptions.FormatError(
                'key "%s" is missing an offset property' % k)
        if not isinstance(offset, int):
            if offset == 'all':
                if len(items) != 1:
                    raise jsonschema.exceptions.FormatError(
                        'length "all" specified but schema length is not 1')
                else:
                    # We have only one offset ("all"), and it is valid. We are
                    # done.
                    return True
            else:
                raise jsonschema.exceptions.FormatError(
                    'offset for key "%s" must be int or "all"' % k)

    items.sort(key=lambda item: item[1]['offset'])

    offset = items[0][1]['offset']
    if offset != 0:
        raise jsonschema.exceptions.FormatError(
            'lowest offset is not 0 for key "%s"' % items[0][0])

    for i in range(1, len(items)-1):
        k, v = items[i]
        offset = v['offset']
        length = v['len']
        if length == 0:
            raise jsonschema.exceptions.FormatError('key %s has 0 length' % k)
        next_offset = items[i+1][1]['offset']
        if offset + length != next_offset:
            next_key = items[i+1][0]
            return ('in key "%s", offset + length overlaps with next key "%s"'
                    % (k, next_key))

    return True



'''
A format checker that checks for strings that are valid infix notation. Four
variables are allowed (A, B, C, D), which stand in for data that will be
manipulated after the schema is parsed. Below are a few valid examples:
A * B
A + 255.0
(A - B) / 10
'''

import ast
import jsonschema
import operator
import simpleeval


FORMAT_NAME = 'infix'


def check(val):
    '''Returns True if val is a valid infix expression and raises FormatError
    otherwise.'''
    operators = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.floordiv,
    }

    # We don't actually care what the names evaluate to, as we're not using
    # SimpleEval() to actually evaluate anything; we're just using it as a
    # shortcut to check whether or not the expression is valid in-fix. So, all
    # the values given here are just dummy placeholders.
    names = {
        'A': 0,
        'B': 0,
        'C': 0,
        'D': 0,
    }
    infix_eval = simpleeval.SimpleEval(operators=operators, names=names)

    try:
        infix_eval.eval(val)
    except:  # noqa: E722
        # eval can throw anything the Python parser can throw, so it's hard to
        # catch every possible item. So, exempt ourselves from this particular
        # flake8 warning.
        raise jsonschema.exceptions.FormatError('"%s" is not valid infix'
                                                % val)

    return True
